/*
hsweep.c Version 0.1.0  Hash sweep database
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/* read the database in physical sequence */
/* one record at a time maintaining position */
/* in hshhdr */
/* hshhdr->ofst   points to the block */

/* valid values for position: */
/* SWPFRST first physical record */
/* SWPNEXT next  physical record */

#include <bthash.h>

int hsweep(hashfmt *hshhdr, int position,
   unsigned char *buf)
   {
   BOOL  bResult;
   DWORD msgerr;
   LPVOID lpMsgBuf;
   DWORD rdlen;
   long loc;
   long hashend;
   unsigned char *p,*q,*r;
   hshhdr->reason = 0;
   hashend = (long) hshhdr->hashend;
   if (buf != NULL)
      {
      p = (unsigned char *) buf;
      q = (unsigned char *) buf + hshhdr->rcdlen;
      while (p < q) *p++ = '\0';
      } /* if buf specified */
   if (position == SWPNEXT)
      {
      hshhdr->ofst += hshhdr->rcdnode;
      } /* if next read */
   else if (position == SWPFRST)
      {
      hshhdr->seq  = 0;
      hshhdr->ofst = HDRLEN;
      loc = hseek(hshhdr,HDRLEN);
      if (hshhdr->reason) return(loc);
      } /* if first read */
   else
      {
      fprintf(stderr,"hsweep: invalid "
	 "position %d\n", position);
      hshhdr->reason = HASH_ERR + 2301;
      return(SWPERR);
      } /* if first read */
   /* for each record in database */
   while (1)
      {
      bResult = ReadFile(hshhdr->hndl,
         (unsigned char *) hshhdr->node,
         (DWORD) hshhdr->rcdnode,
         &rdlen, NULL);
      if (!bResult)
         {
         msgerr = GetLastError();
         hshhdr->reason = SYS_ERR + msgerr;
         FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
            FORMAT_MESSAGE_FROM_SYSTEM |
            FORMAT_MESSAGE_IGNORE_INSERTS,
            NULL,           /* source */
            msgerr,
            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            (LPTSTR) &lpMsgBuf,
            0,              /* msglen */
            NULL);          /* insertion values */
         fprintf(stderr,"hsweep: %s\n", lpMsgBuf);
         LocalFree(lpMsgBuf);
         fprintf(stderr,"hsweep: %s: read error\n",
            hshhdr->dbname);
         hshhdr->reason = SYS_ERR + msgerr;
         return(SWPERR);
         } /* if read err */
      if (rdlen == (DWORD) hshhdr->rcdnode)
	 {
         if (hshhdr->node->next == UINT_MAX)
	    {
            hshhdr->ofst += hshhdr->rcdnode;
            continue;
	    } /* empty record: read another */
	 } /* if good read */
      else if (!rdlen)
         return(SWPEOF);
      else
         {
         fprintf(stderr,"hsweep: %s: read error\n",
            hshhdr->dbname);
         fprintf(stderr,"hsweep: read length %d\n",
            rdlen);
         hshhdr->reason = HASH_ERR + 2302;
         return(SWPERR);
         } /* if read length err */
      p = (unsigned char *) hshhdr->key;
      q = (unsigned char *) p + hshhdr->keylen;
      while (p < q)
	 {
	 if (*p != 0xff) break;
	 p++;
	 } /* test for all ff */
      /* if record key is all ff */
      /* it is a free block */
      /* read next record */
      if (p >= q)
	 {
         hshhdr->ofst += hshhdr->rcdnode;
         continue;
	 } /* if deleted overflow record */
      if (hshhdr->node->next != UINT_MAX)
         {
	 hshhdr->seq++;
         break;
         } /* record found */
      } /* read until record found */
   p = (unsigned char *) hshhdr->key;
   q = (unsigned char *) p + hshhdr->keylen;
   while (p < q)
      {
      if (*p != 0x00) break;
      p++;
      } /* for each byte in key */
   /* if not deleted or empty */
   /* deleted records in hashing space */
   /* are cleared to zeros but retain */
   /* next pointer to overflow chain */
   if (p < q)
      {
      hshhdr->hashofst = hash(hshhdr,
         hshhdr->key);
      if (hshhdr->ofst < hashend)
         {
         if (hshhdr->hashofst != hshhdr->ofst)
	    {
	    fprintf(stderr,"hsweep: invalid "
	       "hashing address\n");
            hshhdr->reason = HASH_ERR + 2303;
            return(HASH_ERR + 2303);
	    } /* inconsistent hashing address */
         } /* if in hashing address space */
      } /* if not deleted or empty */
   if (buf != NULL)
      {
      p = (unsigned char *) hshhdr->node;
      q = (unsigned char *) p + hshhdr->rcdlen;
      r = (unsigned char *) buf;
      while (p < q) *r++ = *p++;
      } /* if buffer specified */
   return(SWPFOUND);
   } /* hsweep */
